/*
** Module   : DMOValidator.java
** Abstract : Responsible for all runtime validation on a DMO
**
** Copyright (c) 2004-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-------------------------------
** 001 ECF 20051106   @23392 Created initial version. Performs record- and
**                           field-level validation of database records,
**                           including mandatory and unique constraints,
**                           as well as custom validation processing (as
**                           converted from Progress validation
**                           expressions).
** 002 ECF 20060116   @23944 Fix for unique constraint validation. Ensure
**                           record found is not the current record.
** 003 ECF 20060116   @23949 Fix for not-null validation. Skip full record
**                           check if a single property is to be checked,
**                           which is nullable.
** 004 ECF 20060126   @24351 Fix latent bug exposed by temp table support.
**                           NPE was possible checking not-null constraint
**                           on a property with 'field' access (instead of
**                           get/set methods).
** 005 ECF 20060308   @25013 Added natural join validation support. Allow
**                           validation of DMO properties (instead of only
**                           BaseDataType properties). Improved error
**                           message content.
** 006 ECF 20060608   @27004 Changed unique constraint validation. Fixes a
**                           recent regression and uses a slightly more
**                           efficient query.
** 007 ECF 20060627   @27588 Cleanup. Removed unused local variable.
** 008 ECF 20060713   @28044 Use new DBUtils class.
** 009 ECF 20060719   @28102 Use updated signature for Persistence.list().
** 010 ECF 20061207   @31619 Eliminated redundant database access. Unique
**                           constraint validation was executing the same
**                           query once per column participating in the
**                           unique constraint.
** 011 ECF 20070130   @32038 Fixed validation of temp-table records.
**                           Uniqueness constraint validation was not
**                           taking the multiplex ID of a TemporaryBuffer
**                           into account.
** 012 ECF 20070212   @32153 Integrated dirty reads into unique constraint
**                           validation. Uncommitted transaction data from
**                           other client contexts must be used when
**                           performing this validation, in order to mimic
**                           Progress' behavior.
** 013 ECF 20070416   @33021 Integrated user interrupt handling.
** 014 ECF 20070507   @33457 Fixed ClassCastException in
**                           uniqueConstraintViolation() method. Was not
**                           taking multiplex ID and primary key ID into
**                           account when reporting error.
** 015 ECF 20070523   @33726 Added validate() variants to validate a
**                           different DMO than is currently in a record
**                           buffer.
** 016 ECF 20070531   @33912 Removed DMO-based validation method handling.
**                           This was an artifact of early planning to
**                           implement validation methods in DMO classes.
**                           These ultimately were moved into business
**                           logic, so this processing is no longer
**                           necessary or desirable, since it adds
**                           overhead.
** 017 EVL 20070609   @34005 Adding explicit import of the class
**                           org.hibernate.type.Type to eliminate conflict
**                           with the same class from java.lang.reflect
**                           package to be able to compile for Java 6.
** 018 ECF 20070827   @34999 Optimized certain temp table queries.
**                           Eliminate multiplex ID check from where
**                           where possible and hard code multiplex ID
**                           when this test is necessary, rather than
**                           using a query substitution parameter.
** 019 ECF 20070911   @35286 Replaced database name string with Database
**                           information object. Required for non-local
**                           database connections. Cleaned up error
**                           messages.
** 020 ECF 20080330   @37737 Fixed unique constraint check. The HQL being
**                           generated was not able to use indices for
**                           some tables. Multiplex ID was not being
**                           properly integrated for temp tables. Also,
**                           text property references were not being
**                           properly generated to account for trimming
**                           and case sensitivity. As a result, we were
**                           doing table scans for every unique constraint
**                           check on these tables! Major performance
**                           improvement. Also integrated generics.
** 021 ECF 20080402   @38407 Minor improvement to validateUnique().
**                           Removed rightTrim of (character) query
**                           substitution parameters, since this is
**                           redundant with the parameter preprocessing
**                           Persistence does. This also removes a subtle
**                           defect found by CA (this call was always
**                           converting character args to empty strings).
** 022 CA  20080526   @38392 Fixed ClassCastException caused by the method
**                           uniqueConstraintViolation - the multiplex id
**                           should be skipped, for temporary records.
**                           Ref: #38391
** 023 ECF 20080605   @38601 Rewrote unique constraint validation. It now
**                           uses DirtyShareContext instead of
**                           ValidationContext and DirtyReadManager.
** 024 SVL 20080620   @38882 validateUnique() obtains EXCLUSIVE lock 
**                           on unique indexes from the very beginning
**                           (instead of SHARE at first, then EXCLUSIVE).
** 025 CA  20080716   @39123 Modifications due to DirtyShareContext API 
**                           changes.
** 026 ECF 20080827   @39607 Removed unused validate(Persistable, Method)
**                           method. This originally was intended for the
**                           invocation of custom validation logic, but
**                           that is handled directly in business logic
**                           instead.
** 027 SVL 20080905   @39701 Optimization of the case when we meet an
**                           unknown value in the DMO fields fixed into 
**                           validateUnique(..).
** 028 ECF 20090220   @41352 Removed unused data structure. The messages
**                           map was originally intended to support custom
**                           validation logic, but that is handled in
**                           business logic instead.
** 029 ECF 20090312   @41638 Fixed validateUnique(). It was possible for
**                           this method to skip validation for a new,
**                           transient record which was not unique due to
**                           non-unique default values.
** 030 GES 20090424   @41937 Import change.
** 031 ECF 20090508   @42123 Rewrote key parts of validation processing to
**                           allow more incremental validation. Also moved
**                           most dirty property sharing from RecordBuffer to
**                           this class.
** 032 ECF 20090512   @42151 Fixed regressions caused by #031 (@42123).
** 033 ECF 20090513   @42165 Fixed checkUniqueAndShare(). If lone dirty
**                           property was not indexed and represented an
**                           extent field, it would cause a problem sharing
**                           the update via the DirtyShareManager.
** 034 CA  20090709   @43125 Fixed checkUnique(). During unique constraint
**                           check, the new value is handled same as the old
**                           values (unknown is always unique and value is
**                           upper cased for a case-insensitive character
**                           column).
** 035 ECF 20090714   @43216 Optimized unique constraint checks. Check whether
**                           backing table is empty and skip unique check in
**                           this case. Use dialect-optimized queries against
**                           the dirty database.
** 036 ECF 20090724   @43417 Eliminated unnecessary validation. If no indexes
**                           are affected by a pending change, skip unique
**                           constraint validation and dirty share processing.
** 037 AIL 20121206          Update from character to TextOps.
** 038 VMN 20130711          Fixes in messages in checkNotNull().
** 039 ECF 20131010          Added metadata support.
** 040 ECF 20131028          Import change.
** 041 SVL 20131028          Use legacy table/buffer name in uniqueConstraintViolation.
*/

package com.goldencode.p2j.persist;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import org.apache.commons.logging.*;
import org.hibernate.type.Type;
import com.goldencode.p2j.persist.dirty.*;
import com.goldencode.p2j.persist.lock.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Performs record-level and field-level validation of database records.
 * This class is intended to support the Progress semantic of "real-time"
 * data validation, where assignments of new values to fields of the record
 * held in a record buffer receive immediate validation against database
 * constraints and schema-level validation expressions.  The types of
 * validation are:
 * <ul>
 *   <li>mandatory (non-nullable) constraint checking;
 *   <li>unique constraint checking;
 *   <li>custom validation expression execution.
 * </ul>
 * <p>
 * If validation is triggered for a single field, only the constraints and
 * validation expression (if any) related to that field are validated.  If
 * validation is triggered for the entire record, all constraints are checked
 * and all validation expressions are executed.
 * <p>
 * <strong>Mandatory Field Check</strong><br>
 * This type of check can be done completely in-process, without accessing
 * the database at all.  A single field check is performed within the context
 * of an impending update to the field's current value.  The proposed new
 * value is provided by the caller and the test is simply that this value
 * does not represent the unknown value.  For a full record check, reflection
 * is used to retrieve the current values of all non-nullable properties in a
 * DMO.  If any value (including any element in an array property) represents
 * the unknown value, the check fails.
 * <p>
 * <strong>Unique Constraint Check</strong><br>
 * This type of check requires the execution of (at least) one query against
 * the database per constraint being checked.  This check can be triggered
 * during update of a single field, or arbitrarily against an entire record.
 * To check against a single constraint, reflection is used to collect the
 * current value of each component property (except in the case of an update,
 * in which case the proposed new value is used for the property being
 * updated).  This data is used for the substitution parameters of a
 * pre-defined HQL query.  The query is submitted to the database and if any
 * record returns, the validation fails.  For a full record check, the same
 * process is repeated for all unique constraints defined for the DMO.  In
 * the latter case, <b>all</b> data is collected from the DMO via reflection.
 * Also in the latter case, the modified record must not be flushed to the
 * database before the query is executed (it is the default behavior of
 * Hibernate to flush modifications to the database before executing any
 * query).  Thus, we explicitly disable flushing temporarily in this case,
 * then re-enable it after the query is complete.  In addition, to fully
 * mimic the Progress semantic, we simulate the ability to perform dirty
 * reads across concurrent transactions in other contexts, using the {@link
 * com.goldencode.p2j.persist.dirty.DirtyShareContext DirtyShareContext}.
 * <p>
 * <strong>Custom Validation</strong><br>
 * Not used at this time.  This was originally intended to support field-level
 * validation expressions.  However, these are implemented in converted
 * business logic today because they can invoke arbitrary code paths.
 * <p>
 * <b>Important:</b> while instances of this class contain mapped data to
 * enable lookups of information required for validation, these data
 * structures are all initialized at the time the instance is constructed.
 * Instances of this class are shared across record buffers and across user
 * contexts, so they must remain <i>stateless</i> with respect to the buffers
 * and DMO records being validated.
 */
final class DMOValidator
{
   /** Logger */
   private static final Log LOG = LogFactory.getLog(DMOValidator.class);
   
   /** Cache of previously created instances, mapped by DMO interface */
   private static final Map<Class<?>, DMOValidator> cache =
      new HashMap<Class<?>, DMOValidator>();
   
   /** Map of unique index names to <code>HQLBundles</code> */
   private final Map<String, HQLBundle> uniqueConstraints =
      new HashMap<String, HQLBundle>();
   
   /** Map of non-nullable DMO properties to their getter methods */
   private final Map<String, Method> notNullProperties =
      new HashMap<String, Method>();
   
   /** Names of properties which trigger validation mapped to their getters */
   private final Map<String, Method> triggers =
      new HashMap<String, Method>();
   
   /** Setter methods, mapped by property names */
   private final Map<String, Method> setterMap;
   
   /** Property names, mapped by getter methods */
   private final Map<Method, String> propertyNames =
      new HashMap<Method, String>();
   
   /** Ignore case flags, mapped by character getter methods */
   private Map<Method, Boolean> charProps = null;
   
   /**
    * Construct a validator, based upon the database, and the DMO interface
    * and class represented by <code>buffer</code>.
    * <p>
    * The following data structures are initialized in this constructor:
    * <ul>
    *   <li>{@link #triggers} - A map of DMO properties to their companion
    *       getter methods.  Only those properties whose modification must
    *       trigger a validation check (of any type) are keys in this map.
    *       The getter method values of this map are provided for convenience
    *       during validation processing.
    *   <li>{@link #notNullProperties} - A map of DMO properties to their
    *       companion getter methods.  The contents of this map are a subset
    *       of the contents of the <code>triggers</code> map.  Only those
    *       setters which update non-nullable properties are contained in this
    *       map.
    *   <li>{@link #uniqueConstraints} - A map of unique index names to the
    *       {@link HQLBundle} objects which will be used to issue database
    *       queries to test for the existence of records which should be
    *       unique.
    * </ul>
    *
    * @param   buffer
    *          Record buffer which contains information required to create a
    *          new validator instance.
    */
   private DMOValidator(RecordBuffer buffer)
   {
      Database database = buffer.getDatabase();
      Class<?> dmoIface = buffer.getDMOInterface();
      Class<?> dmoClass = buffer.getDMOImplementationClass();
      
      setterMap = PropertyHelper.settersByProperty(dmoIface);
      Map<String, Method> getterMap =
         PropertyHelper.gettersByProperty(dmoIface);
      
      // Map property names by getter methods (for error messages).
      Iterator<Map.Entry<String, Method>> gbpIter =
         getterMap.entrySet().iterator();
      while (gbpIter.hasNext())
      {
         Map.Entry<String, Method> next = gbpIter.next();
         propertyNames.put(next.getValue(), next.getKey());
      }
      
      // Collect all DMO setter methods which should trigger a not-null test.
      Iterator<String> nnIter = DatabaseManager.notNullProperties(database,
                                                                  dmoClass);
      while (nnIter.hasNext())
      {
         String name = nnIter.next();
         Method getter = getterMap.get(name);
         if (getter != null) // could have "field" access
         {
            notNullProperties.put(name, getter);
            triggers.put(name, getter);
         }
      }
      
      // Map unique index names for the target DMO type to HQLBundles which
      // will be used when performing unique constraint validation.
      IndexHelper idxHlp = buffer.getIndexHelper();
      String entity = buffer.getEntityName();
      try
      {
         Map<String, List<String>> map = idxHlp.getIndexMapForEntity(entity,
                                                                     true);
         Iterator<Map.Entry<String, List<String>>> uniqIter =
            map.entrySet().iterator();
         while (uniqIter.hasNext())
         {
            Map.Entry<String, List<String>> next = uniqIter.next();
            String index = next.getKey();
            List<String> properties = next.getValue();
            buildUniqueBundle(index, properties, getterMap, buffer);
         }
      }
      catch (PersistenceException exc)
      {
         throw new StopConditionException(exc);
      }
   }
   
   /**
    * Get an instance of a validator, based upon the DMO interface represented
    * by <code>buffer</code>.  Validator instances are immutable once created,
    * and as such they are shared across multiple buffers and multiple user
    * contexts (they are stateless, so this is thread-safe).
    *
    * @param   buffer
    *          Record buffer which contains information required to create a
    *          new validator instance.
    *
    * @return  A validator instance for use by <code>buffer</code>.
    */
   static DMOValidator getInstance(RecordBuffer buffer)
   {
      synchronized (cache)
      {
         Class<?> dmoIface = buffer.getDMOInterface();
         DMOValidator validator = cache.get(dmoIface);
         if (validator == null)
         {
            validator = new DMOValidator(buffer);
            cache.put(dmoIface, validator);
         }
         
         return validator;
      }
   }
   
   /**
    * Get the set of names of properties which are non-nullable, for the DMO
    * type managed by this validator.
    * 
    * @return  Set of non-nullable property names, or the empty set if there
    *          are no such properties.  If non-empty, the caller can modify
    *          this set, since it is a copy of internal data.
    */
   Set<String> getNonNullables()
   {
      Set<String> set = notNullProperties.keySet();
      if (set.isEmpty())
      {
         return Collections.emptySet();
      }
      
      return (new HashSet<String>(set));
   }
   
   /**
    * Execute validation on the properties and constraints of the target
    * record specified in the given validation parameters.  Assumes all
    * property values to be tested are already set into the object.
    *
    * @param   buffer
    *          Record buffer.
    * @param   target
    *          Record to be validated.  If <code>null</code>, then
    *          <code>buffer</code>'s current record is validated.
    * @param   vData
    *          Validation parameters.
    *
    * @throws  ValidationException
    *          if any aspect of validation fails.
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  RuntimeException
    *          if there is an error reflectively inspecting the record being
    *          validated.
    */
   void validate(RecordBuffer buffer,
                 Persistable target,
                 ValidationData vData)
   throws ValidationException,
          PersistenceException
   {
      validate(buffer, target, vData, null, null);
   }
   
   /**
    * Execute validation on the proposed property value change to the record.
    * Additional property values may be queried from the record to complete
    * the validation.
    * <p>
    * This method is invoked when a setter method is about to be invoked
    * upon a DMO to change the value of one of its properties, or when the
    * entire record must be inspected and validated.  Validation is skipped
    * if <code>property</code> has not been determined to be a validation
    * trigger.
    *
    * @param   buffer
    *          Record buffer containing the record to be validated.
    * @param   vData
    *          Validation parameters.
    * @param   property
    *          DMO property to be updated in <code>target</code> if validation
    *          completes successfully.  If <code>null</code>, the validation
    *          parameters are consulted to determine which properties and
    *          constraints are to be validated.
    * @param   value
    *          Value to be set into <code>property</code> if validation
    *          completes successfully.
    *
    * @throws  ValidationException
    *          if any aspect of validation fails.
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  RuntimeException
    *          if there is an error reflectively inspecting the record being
    *          validated.
    */
   void validate(RecordBuffer buffer,
                 ValidationData vData,
                 String property,
                 BaseDataType value)
   throws ValidationException,
          PersistenceException
   {
      validate(buffer, buffer.getCurrentRecord(), vData, property, value);
   }
   
   /**
    * Execute validation on the proposed property value change to the record.
    * Additional property values may be queried from the record to complete
    * the validation.
    * <p>
    * This method is invoked when a setter method is about to be invoked
    * upon a DMO to change the value of one of its properties, or when the
    * entire record must be inspected and validated.  Validation is skipped
    * if <code>setter</code> would update a property which has not been
    * determined to be a validation trigger.
    *
    * @param   buffer
    *          Record buffer.
    * @param   target
    *          Record to be validated.  If <code>null</code>, then
    *          <code>buffer</code>'s current record is validated.
    * @param   vData
    *          Validation parameters.
    * @param   property
    *          DMO property to be updated in <code>target</code> if validation
    *          completes successfully.  If <code>null</code>, the validation
    *          parameters are consulted to determine which properties and
    *          constraints are to be validated.
    * @param   value
    *          Value to be set into <code>property</code> if validation
    *          completes successfully.
    *
    * @throws  ValidationException
    *          if any aspect of validation fails.
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  RuntimeException
    *          if there is an error reflectively inspecting the record being
    *          validated.
    */
   private void validate(RecordBuffer buffer,
                         Persistable target,
                         ValidationData vData,
                         String property,
                         BaseDataType value)
   throws ValidationException,
          PersistenceException
   {
      try
      {
         if (target == null)
         {
            target = buffer.getCurrentRecord();
         }
         checkNotNull(target, vData, property, value);
         checkUniqueAndShare(buffer, target, vData, property, value);
      }
      catch (IllegalAccessException exc)
      {
         throw new RuntimeException("Error inspecting DMO", exc);
      }
      catch (InvocationTargetException exc)
      {
         DBUtils.handleException(buffer.getDatabase(), exc);
         
         throw new RuntimeException("Error inspecting DMO", exc);
      }
   }
   
   /**
    * Validate a single not-null constraint, or all such constraints defined
    * for the <code>target</code> DMO in the validation parameters, depending
    * upon the value of the given <code>property</code>.
    *
    * @param   target
    *          The DMO to be validated.
    * @param   vData
    *          Validation parameters.
    * @param   property
    *          DMO property to be updated in <code>target</code> if validation
    *          completes successfully.  If <code>null</code>, the validation
    *          parameters are consulted to determine which properties and
    *          constraints are to be validated.
    * @param   value
    *          Value to be set into <code>property</code> if validation
    *          completes successfully.
    *
    * @throws  ValidationException
    *          if a constrained property contains the unknown value.
    * @throws  IllegalAccessException
    *          if a getter method used to inspect the DMO cannot be accessed.
    * @throws  InvocationTargetException
    *          if a getter method used to inspect the DMO throws an exception
    *          upon invocation.
    */
   private void checkNotNull(Persistable target,
                             ValidationData vData,
                             String property,
                             BaseDataType value)
   throws ValidationException,
          IllegalAccessException,
          InvocationTargetException
   {
      if (notNullProperties.isEmpty())
      {
         // Nothing to do!
         return;
      }
      
      Set<String> dirtyProps = vData.getDirtyProperties();
      
      // Single property check.
      if (property != null)
      {
         if (dirtyProps.contains(property) &&
             notNullProperties.containsKey(property) &&
             value.isUnknown())
         {
            throw new ValidationException(
               StringHelper.changeCase(property, true) +
               " is mandatory, but has a value of ?",
               275);
         }
         
         return;
      }
      
      // Walk each non-nullable, dirty property and invoke its getter method;
      // test the result to ensure it is not the unknown value.
      Iterator<String> iter = dirtyProps.iterator();
      while (iter.hasNext())
      {
         String dProp = iter.next();
         Method getter = notNullProperties.get(dProp);
         
         if (getter == null)
         {
            // Dirty property is nullable, no need to check.
            continue;
         }
         
         boolean indexed = (getter.getParameterTypes().length == 1);
         try
         {
            // Invoke simple getter once or indexed getter until out of valid
            // index values.
            for (int i = 0; indexed || i == 0; i++)
            {
               Integer index = (indexed ? new Integer(i) : null);
               Object pv = getPropertyValue(target, getter, index);
               if (pv instanceof BaseDataType)  // TODO:  necessary check?
               {
                  BaseDataType bdt = (BaseDataType) pv;
                  if (bdt.isUnknown())
                  {
                     throw new ValidationException(
                        StringHelper.changeCase(dProp, true) +
                        " is mandatory, but has a value of ?",
                        275);
                  }
               }
               else if (pv == null)
               {
                  throw new ValidationException(
                     "Non-nullable property " +
                     dProp +
                     " is null",
                     -1);
               }
            }
         }
         catch (ArrayIndexOutOfBoundsException exc)
         {
            // TODO:  use size API in DMO interface to detect maximum extent
            // for an indexed property, instead of using an out of bounds
            // exception to change control flow.  This try-catch is a hack.
         }
      }
   }
   
   /**
    * Validate a single unique constraint or all unique constraints defined
    * in the validation parameters for the target record.  This method in most
    * cases must access the database to complete the constraint check.
    *
    * @param   buffer
    *          Record buffer containing the record to be validated.
    * @param   target
    *          The DMO to be validated.
    * @param   vData
    *          Validation parameters.
    * @param   property
    *          DMO property to be updated in <code>target</code> if validation
    *          completes successfully.  If <code>null</code>, the validation
    *          parameters are consulted to determine which properties and
    *          constraints are to be validated.
    * @param   value
    *          Value to be set into <code>property</code> if validation
    *          completes successfully.
    *
    * @throws  ValidationException
    *          if the constrained properties contains values which are not
    *          unique to this record.
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  IllegalAccessException
    *          if a getter method used to inspect the DMO cannot be accessed.
    * @throws  InvocationTargetException
    *          if a getter method used to inspect the DMO throws an exception
    *          upon invocation.
    */
   private void checkUniqueAndShare(RecordBuffer buffer,
                                    Persistable target,
                                    ValidationData vData,
                                    String property,
                                    BaseDataType value)
   throws ValidationException,
          PersistenceException,
          IllegalAccessException,
          InvocationTargetException
   {
      Collection<String> dirtyUnique = vData.getDirtyUnique();
      Collection<String> dirtyNonUnique = vData.getDirtyNonUnique();
      
      if (dirtyUnique.isEmpty() &&
          (dirtyNonUnique.isEmpty() || !vData.isShareDirty()))
      {
         // Nothing to do.
         return;
      }
      
      Persistence persistence = buffer.getPersistence();
      Method omit = (property != null ? triggers.get(property) : null);
      DirtyShareContext dirtyContext = buffer.getDirtyContext();
      String entity = buffer.getEntityName();
      
      try
      {
         // Lock all indexes while validating.
         // TODO:  optimization:  lock only dirty indexes unless we're
         // inserting a new record.  As part of this optimization, we need to
         // consider the remote server case, such that we minimize the data
         // sent across (an array or collections of index name strings may not
         // be efficient).
         dirtyContext.lockAllIndexes(entity, LockType.EXCLUSIVE);

         // If we are validating a record already updated in memory, we must
         // disable flushing before executing validation queries, in case the
         // record would cause a unique constraint violation when flushed to
         // the database immediately before a query.
         if (omit == null)
         {
            persistence.enableFlush(false);
         }

         // Validate each dirty, unique index.
         for (String index : dirtyUnique)
         {
            HQLBundle bundle = uniqueConstraints.get(index);
            if (bundle != null)
            {
               checkUnique(buffer,
                           target,
                           vData,
                           bundle,
                           omit,
                           property,
                           value);
            }
         }
         
         // Share dirty updates if required.
         // NOTE:  currently, it is assumed that ValidationData.isShareDirty()
         // will only return true for the buffer's current record, because the
         // logic below operates on the buffer's current record and current
         // dirty property state.
         boolean share =
            (vData.isShareDirty() &&
             (!dirtyUnique.isEmpty() || !dirtyNonUnique.isEmpty()));
         if (share)
         {
            // Store the update into the dirty database, either as a new
            // insert, or as an update to an existing record.
            if (vData.isInsertRequired())
            {
               Persistable dmo = buffer.getCurrentRecord();
               Method setter = (property != null
                                ? setterMap.get(property)
                                : null);
               dirtyContext.insert(buffer, dmo, setter, value, false);
            }
            else
            {
               boolean propIndexed = (property != null &&
                                      buffer.isPropertyIndexed(property));
               Map<String, Object> dirtyPropMap = buffer.getDirtyProperties();
               int size = dirtyPropMap.size();
               if (size > 0 || propIndexed)
               {
                  int start = 0;
                  if (propIndexed && !dirtyPropMap.containsKey(property))
                  {
                     size++;
                     start++;
                  }
                  
                  String[] dup = new String[size];
                  BaseDataType[] duv = new BaseDataType[size];
                  if (start > 0)
                  {
                     dup[0] = property;
                     duv[0] = value;
                  }
                  
                  Iterator<Map.Entry<String, Object>> dirtyIter =
                     dirtyPropMap.entrySet().iterator();
                  for (int i = start; dirtyIter.hasNext(); i++)
                  {
                     Map.Entry<String, Object> entry = dirtyIter.next();
                     dup[i] = entry.getKey();
                     Object[] diffs = (Object[]) entry.getValue();
                     duv[i] = (BaseDataType) diffs[1];
                  }
                  
                  dirtyContext.update(buffer, dup, duv, false);
               }
            }
         }
         
         // There will be no unique constraint violation as a result of the
         // in-memory change, so flush the updated record (and other pending
         // updates) now, if flushing was deferred earlier. OTOH, if the
         // target record is transient, there is no need to flush the session,
         // because the session is not yet tracking the new record. In this
         // case, defer the flush.
         if (omit == null &&
             !vData.getDirtyProperties().isEmpty() &&
             !vData.isTransient())
         {
            persistence.flush();
         }
      }
      finally
      {
         if (omit == null)
         {
            persistence.enableFlush(true);
         }
         
         // Unlock unique indexes.  See TODO above.
         dirtyContext.lockAllIndexes(entity, LockType.NONE);
      }
   }
   
   /**
    * Validate a single unique constraint defined for the target record.
    * This method in most cases must access the database to complete the
    * constraint check.  However, in cases where one of the constrained
    * values represents the unknown value, the validation completes
    * successfully.  In this case it is not necessary to access the database,
    * since the unknown value is always considered unique.
    *
    * @param   buffer
    *          Record buffer containing the record to be validated.
    * @param   target
    *          The DMO to be validated.
    * @param   vData
    *          Validation parameters.
    * @param   bundle
    *          An object containing the HQL query statement and all getter
    *          methods necessary to gather substitution parameter values from
    *          <code>target</code>.
    * @param   omit
    *          Getter method associated with a property for which we already
    *          have a value.  This method will <i>not</i> be used to retrieve
    *          a value from <code>target</code> via reflection.  Instead,
    *          the <code>value</code> parameter is used.  If
    *          <code>null</code>, all constrained properties are inspected
    *          for current values via reflection.
    * @param   property
    *          DMO property to be updated in <code>target</code> if validation
    *          completes successfully.  If <code>null</code>, the validation
    *          parameters are consulted to determine which properties and
    *          constraints are to be validated.
    * @param   value
    *          Value to be set into <code>property</code> if validation
    *          completes successfully.
    *
    * @throws  ValidationException
    *          if the constrained properties contains values which are not
    *          unique to this record.
    * @throws  PersistenceException
    *          if any error occurs querying the database.
    * @throws  IllegalAccessException
    *          if a getter method used to inspect the DMO cannot be accessed.
    * @throws  InvocationTargetException
    *          if a getter method used to inspect the DMO throws an exception
    *          upon invocation.
    */
   private void checkUnique(RecordBuffer buffer,
                            Persistable target,
                            ValidationData vData,
                            HQLBundle bundle,
                            Method omit,
                            String property,
                            BaseDataType value)
   throws ValidationException,
          PersistenceException,
          IllegalAccessException,
          InvocationTargetException
   {
      // An unknown value short-circuits the uniqueness check, since unknown
      // value is by definition unique.
      if (value != null && value.isUnknown())
      {
         return;
      }
      
      // If there are no records in the backing table, it is impossible for
      // the target record to violate a unique constraint, so we can short-
      // circuit the check.  This potentially preempts a lot of reflection
      // calls and a database query.
      if (buffer.isTableDefinitelyEmpty())
      {
         return;
      }
      
      Persistence persistence = buffer.getPersistence();
      
      // Gather substitution parameters.
      Object[] args = null;
      
      // First substitution parameter for a temp table is the multiplex ID.
      boolean temporary = buffer.isTemporary();
      if (temporary)
      {
         args = new Object[bundle.gettersSize() + 1];
         args[0] = buffer.getMultiplexID();
      }
      else
      {
         args = new Object[bundle.gettersSize()];
      }
      
      // Loop through properties in the unique constraint and populate the
      // parameter array.
      boolean forceValid = false;
      Iterator<Method> iter = bundle.getters();
      for (int i = (temporary ? 1 : 0); iter.hasNext(); i++)
      {
         Method getter = iter.next();
         BaseDataType bdt = null;

         if (getter.equals(omit))
         {
            // Updated value is not in the DMO yet, so we need to use the
            // value provided to this method.
            bdt = value;
         }
         else
         {
            // Get current value of constraint component in DMO.
            bdt = (BaseDataType) getter.invoke(target);
         }
         
         // An unknown value short-circuits the check, since unknown value
         // is by definition unique.
         if (bdt.isUnknown())
         {
            forceValid = true;
            break;
         }
         
         if (charProps != null && bdt instanceof character)
         {
            // Possibly uppercase text parameter before passing it to query.
            character c = (character) bdt;
            Boolean ignoreCase = charProps.get(getter);
            if (ignoreCase != null && ignoreCase)
            {
               c = TextOps.toUpperCase(c);
            }
            args[i] = c;
         }
         else
         {
            args[i] = bdt;
         }
      }
      
      DirtyShareContext dirtyContext = buffer.getDirtyContext();
      String entity = buffer.getEntityName();
      
      if (!forceValid)
      {
         Type[] types = DBUtils.makeTypeArray(args);
         
         // Validate against primary data.
         String hql = bundle.getStatement(0, false);
         List results = persistence.list(hql, args, types, 1, 0, true);
         Serializable candidateID = target.getId();
         if (results != null)
         {
            // A record was found.  Make sure it's not the current record,
            // and make sure a modified version does not invalidate the
            // violation.
            Serializable foundID = (Serializable) results.get(0);
            if (!candidateID.equals(foundID) &&
                doubleCheckUniqueViolation(foundID,
                                           buffer,
                                           bundle,
                                           args,
                                           true))
            {
               uniqueConstraintViolation(buffer, bundle, args);
            }
         }
      
         // Validate against uncommitted data.
         hql = bundle.getStatement(0, true);
         results = dirtyContext.list(entity, hql, args, types, 1, 0, true);
         if (results != null)
         {
            // A record was found.  Make sure it's not the current record,
            // and make sure a modified version does not invalidate the
            // violation.
            Serializable foundID = (Serializable) results.get(0);
            if (!candidateID.equals(foundID) &&
                doubleCheckUniqueViolation(foundID,
                                           buffer,
                                           bundle,
                                           args,
                                           false))
            {
               uniqueConstraintViolation(buffer, bundle, args);
            }
         }
      }
   }
   
   /**
    * Double check whether a DMO which has been found as the result of a
    * unique constraint check lookup actually violates the constraint.  The
    * lookup can produce a false positive if there is a dirty version of the
    * record (either in the dirty share manager or in the current context's
    * Hibernate session) which has been modified in such a way so as to make
    * the found record no longer unique.
    * <p>
    * In this method, we find that dirty record (if any) and test its relevant
    * properties to confirm or refute the unique constraint violation.
    * 
    * @param   foundID
    *          ID of the record which was found to violate the constraint.
    * @param   buffer
    *          Record buffer associated with the record under test.
    * @param   bundle
    *          Bundle containing the getter methods we will use to test the
    *          relevant properties if a dirty version of the suspect record is
    *          found.
    * @param   args
    *          Property values which are being tested for uniqueness.
    * @param   primary
    *          <code>true</code> if suspect record was found in the primary
    *          database;  <code>false</code> if suspect record was found in
    *          the dirty share manager.
    * 
    * @return  <code>true</code> if the violation is confirmed, else
    *          <code>false</code>.
    * @throws  PersistenceException
    *          if there are any database errors performing the confirmation.
    */
   private boolean doubleCheckUniqueViolation(Serializable foundID,
                                              RecordBuffer buffer,
                                              HQLBundle bundle,
                                              Object[] args,
                                              boolean primary)
   throws PersistenceException
   {
      /*
      if (primary)
      {
         return true;
      }
      
      boolean evict = false;
      String entity = buffer.getEntityName();
      Persistence persistence = buffer.getPersistence();
      RecordIdentifier ident = new RecordIdentifier(entity, foundID);
      Persistable found = persistence.quickLoad(ident);
      
      // If record found in dirty share manager no longer exists in our
      // version of the primary database, we can assume we deleted it earlier
      // in this session.
      if (found == null)
      {
         return false;
      }
      else
      {
         buffer.evictDMOIfUnused(found);
      }
      */
      
      /*
      boolean evict = false;
      String entity = buffer.getEntityName();
      Persistence persistence = buffer.getPersistence();
      DirtyShareContext dirtyContext = buffer.getDirtyContext();
      Persistable found = null;
      
      if (primary)
      {
         found = dirtyContext.getDirtyDMO(entity, foundID);
      }
      
      if (found == null)
      {
         RecordIdentifier ident = new RecordIdentifier(entity, foundID);
         found = persistence.quickLoad(ident);
         evict = true;
      }
      
      if (found != null)
      {
         try
         {
            Iterator<Method> iter = bundle.getters();
            for (int i = 0; iter.hasNext(); i++)
            {
               Method getter = iter.next();
               
               try
               {
                  Object value = getter.invoke(found);
                  if (!args[i].equals(value))
                  {
                     return false;
                  }
               }
               catch (Exception exc)
               {
                  LOG.error("Reflection error", exc);
                  
                  break;
               }
            }
         }
         finally
         {
            if (evict)
            {
               buffer.evictDMOIfUnused(found);
            }
         }
      }
      */
      
      return true;
   }
   
   /**
    * Throw a validation exception for a unique constraint violation.  An
    * error message is composed describing the constraint properties and
    * their current values.
    *
    * @param   buffer
    *          Record buffer representing the DMO whose constraint has been
    *          violated.
    * @param   bundle
    *          Object from  which property names will be extracted.
    * @param   args
    *          Current values of the properties which participate in the
    *          unique constraint.
    *
    * @throws  ValidationException
    *          always.
    */
   private void uniqueConstraintViolation(RecordBuffer buffer,
                                          HQLBundle bundle,
                                          Object[] args)
   throws ValidationException
   {
      BufferImpl dmo = (BufferImpl) buffer.getDMOProxy();
      String legacyName = buffer.isDynamic() ? TableMapper.getLegacyName(dmo)
                                             : dmo.getName().getValue();

      StringBuilder buf = new StringBuilder();
      buf.append(legacyName);
      buf.append(" already exists with");
      
      // Skip the multiplex id for the temporary buffer
      int i = (buffer.isTemporary() ? 1 : 0);
      for (Iterator<Method> iter = bundle.getters(); iter.hasNext(); i++)
      {
         Method getter = iter.next();
         String propName = propertyNames.get(getter);
         if ("id".equals(propName))
         {
            continue;
         }
         buf.append(" ");
         // TODO use legacy field name
         buf.append(propName);
         buf.append(" ");
         buf.append(((BaseDataType) args[i]).toStringMessage());
      }
      
      throw new ValidationException(buf.toString(), 132);
   }
   
   /**
    * Construct and populate an <code>HQLBundle</code> suitable for
    * producing a query to test a unique constraint.  The bundle contains a
    * single HQL statement.  The statement's where clause contains an
    * equality match for each property contained in <code>names</code>.
    * It selects the unique record ID from the DMO currently associated with
    * <code>buffer</code>.  For example:
    * <pre>
    *   select customer.id from CustomerImpl as customer
    *   where customer.prop1 = ? and customer.prop2 = ?
    * </pre>
    * <p>
    * A getter method is added to the bundle for the property associated with
    * each equality match in the where clause.  The bundle is stored in the
    * map of unique constraints, keyed by the setter method associated with
    * each property component of the unique constraint.
    * 
    * @param   index
    *          Name of the unique index which defines the constraint to be
    *          validated.
    * @param   properties
    *          List of property names which comprise a unique constraint for
    *          table represented by the given record buffer.
    * @param   getters
    *          Map of getter methods by property names for the DMO interface
    *          associated with the given record buffer.
    * @param   buffer
    *          Record buffer for which this HQLBundle is to be generated.
    * 
    * @throws  PersistenceException
    *          if any errors occur looking up metadata regarding a property.
    */
   private void buildUniqueBundle(String index,
                                  List<String> properties,
                                  Map<String, Method> getters,
                                  RecordBuffer buffer)
   throws PersistenceException
   {
      HQLBundle bundle = new HQLBundle();
      
      String alias = buffer.getDMOAlias();
      StringBuilder buf = new StringBuilder();
      buf.append("select ");
      buf.append(alias);
      buf.append(".id from ");
      buf.append(buffer.getDMOImplementationName());
      buf.append(" as ");
      buf.append(alias);
      
      boolean temporary = buffer.isTemporary();
      int start = 0;
      if (temporary)
      {
         buf.append(" where ");
         buf.append(alias);
         buf.append(".");
         buf.append(TemporaryBuffer.MULTIPLEX_FIELD_NAME);
         buf.append(" = ?");
         start++;
      }
      
      StringBuilder bufDirty = (temporary ? null : new StringBuilder(buf));
      Class<?> dmoClass = buffer.getDMOImplementationClass();
      Iterator<String> iter = properties.iterator();
      for (int i = start; iter.hasNext(); i++)
      {
         String property = iter.next();
         
         buf.append(i == 0 ? " where " : " and ");
         DBUtils.composePropertyName(buffer, property, buf, Database.Type.PRIMARY);
         buf.append(" = ?");
         
         if (!temporary)
         {
            bufDirty.append(i == 0 ? " where " : " and ");
            DBUtils.composePropertyName(buffer, property, bufDirty, Database.Type.DIRTY);
            bufDirty.append(" = ?");
         }
         
         Method getter = getters.get(property);
         
         bundle.addGetter(getter);
         
         Boolean ignoreCase = DatabaseManager.getIgnoreCase(dmoClass,
                                                            property);
         if (ignoreCase != null)
         {
            if (charProps == null)
            {
               charProps = new HashMap<Method, Boolean>(3);
            }
            charProps.put(getter, ignoreCase);
         }
         
         triggers.put(property, getter);
      }
      
      String primaryHQL = buf.toString();
      bundle.addStatement(primaryHQL, false);
      
      String dirtyHQL = (temporary ? primaryHQL : bufDirty.toString());
      bundle.addStatement(dirtyHQL, true);
      
      uniqueConstraints.put(index, bundle);
      
      if (LOG.isDebugEnabled())
      {
         StringBuilder sb = new StringBuilder("Unique validation bundle:\n");
         sb.append(bundle);
         LOG.debug(sb.toString());
      }
   }
   
   /**
    * Extract current value of a DMO property, given the DMO, the getter
    * method for the property, and an optional index argument to the getter
    * method.  Uses reflection to obtain the property value.
    *
    * @param   target
    *          Target DMO record.
    * @param   getter
    *          Getter method for the property.
    * @param   index
    *          Index value to pass to the getter (for indexed getters only),
    *          or <code>null</code> if getter takes no arguments.
    *
    * @return  Current value of the property.
    *
    * @throws  IllegalAccessException
    *          if getter method used to inspect the DMO cannot be accessed.
    * @throws  InvocationTargetException
    *          if getter method used to inspect the DMO throws an exception
    *          upon invocation.
    */
   private Object getPropertyValue(Object target, Method getter, Object index)
   throws IllegalAccessException,
          InvocationTargetException
   {
      Object value = null;
      if (index != null)
      {
         Object[] args = { index };
         value = getter.invoke(target, args);
      }
      else
      {
         value = getter.invoke(target);
      }
      
      return value;
   }
}
